import uuid
from typing import Any, Optional, Union

from sqlalchemy.dialects import postgresql
from sqlalchemy.engine.interfaces import Dialect
from sqlalchemy.types import CHAR, TypeDecorator


class UUIDType(TypeDecorator):  # type: ignore
    """Backend-agnostic UUID type."""
    impl = CHAR
    cache_ok = True

    def load_dialect_impl(self, dialect: Dialect) -> Any:
        if dialect.name == 'postgresql':
            return dialect.type_descriptor(postgresql.UUID())
        return dialect.type_descriptor(CHAR(32))

    def process_bind_param(
        self,
        value: Union[int, str, uuid.UUID],
        dialect: Dialect,
    ) -> Optional[str]:
        if value is None:
            return value
        if dialect.name == 'postgresql':
            return str(value)
        if isinstance(value, str):
            value = uuid.UUID(value)
        if isinstance(value, uuid.UUID):
            return f'{value.int:0>32x}'
        raise TypeError

    def process_result_value(
        self,
        value: Optional[Union[str, uuid.UUID]],
        dialect: 'Dialect',
    ) -> 'Optional[uuid.UUID]':
        if value is None:
            return value
        if not isinstance(value, uuid.UUID):
            value = uuid.UUID(value)
        return value



from datetime import datetime
from typing import Any, Optional
from uuid import uuid4

import sqlalchemy as sa
from sqlalchemy import orm

from sqlalchemy_things.column_types import UUIDType
from sqlalchemy_things.declarative.base import (
    DeclarativeMixin,
    InheritedDeclarativeMixin,
    get_inherited_column,
)


@orm.declarative_mixin
class BigIntegerPrimaryKeyMixin(DeclarativeMixin):
    """
    Big integer primary key mixin.

    .. warning::
      SQLite backend not support autoincrement for BigInteger column type.
    """
    @orm.declared_attr  # type: ignore
    def pk(cls) -> sa.Column:
        default = sa.Column(sa.BigInteger, primary_key=True)
        get_inherited_column(cls, 'pk', default)
        return default


@orm.declarative_mixin
class CascadeBigIntegerPrimaryKeyMixin(InheritedDeclarativeMixin):
    """
    Cascade big integer primary key mixin for joined table inheritance.

    .. warning::
      SQLite backend not support autoincrement for BigInteger column type.
    """
    @orm.declared_attr.cascading  # type: ignore
    def pk(cls) -> sa.Column:
        if orm.has_inherited_table(cls) is False:
            return sa.Column(sa.BigInteger, primary_key=True)
        return get_inherited_primary_key(cls)


@orm.declarative_mixin
class CascadeDateTimePrimaryKeyMixin(InheritedDeclarativeMixin):
    """Cascade datetime primary key mixin for joined table inheritance."""
    @orm.declared_attr.cascading  # type: ignore
    def pk(cls) -> sa.Column:
        if orm.has_inherited_table(cls) is False:
            return sa.Column(sa.DateTime, primary_key=True,
                             default=datetime.now)
        return get_inherited_primary_key(cls)


@orm.declarative_mixin
class CascadeIntegerPrimaryKeyMixin(InheritedDeclarativeMixin):
    """Cascade integer primary key mixin for joined table inheritance."""
    @orm.declared_attr.cascading  # type: ignore
    def pk(cls) -> sa.Column:
        if orm.has_inherited_table(cls) is False:
            return sa.Column(sa.Integer, primary_key=True)
        return get_inherited_primary_key(cls)


@orm.declarative_mixin
class CascadeUUIDPrimaryKeyMixin(InheritedDeclarativeMixin):
    """Cascade UUID primary key mixin for joined table inheritance."""
    @orm.declared_attr.cascading  # type: ignore
    def pk(cls) -> sa.Column:
        if orm.has_inherited_table(cls) is False:
            return sa.Column(UUIDType, primary_key=True, default=uuid4)
        return get_inherited_primary_key(cls)


@orm.declarative_mixin
class DateTimePrimaryKeyMixin(DeclarativeMixin):
    """Datetime primary key mixin."""
    @orm.declared_attr  # type: ignore
    def pk(cls) -> sa.Column:
        default = sa.Column(sa.DateTime, primary_key=True, default=datetime.now)
        get_inherited_column(cls, 'pk', default)
        return default


@orm.declarative_mixin
class IntegerPrimaryKeyMixin(DeclarativeMixin):
    """Integer primary key mixin."""
    @orm.declared_attr  # type: ignore
    def pk(cls) -> sa.Column:
        default = sa.Column(sa.Integer, primary_key=True)
        get_inherited_column(cls, 'pk', default)
        return default


@orm.declarative_mixin
class ParentPrimaryKeyMixin(InheritedDeclarativeMixin):
    """
    Parent primary key mixin for child class in joined table inheritance without
    cascade primary key mixin.

    .. warning::
      This mixin must precede the parent class when declaring a child class.
    """
    @orm.declared_attr  # type: ignore
    def pk(cls) -> sa.Column:
        return get_inherited_primary_key(cls)


@orm.declarative_mixin
class UUIDPrimaryKeyMixin(DeclarativeMixin):
    """UUID primary key mixin."""
    @orm.declared_attr  # type: ignore
    def pk(cls) -> sa.Column:
        default = sa.Column(UUIDType, primary_key=True, default=uuid4)
        get_inherited_column(cls, 'pk', default)
        return default


def get_inherited_primary_key(cls: Any) -> Optional[sa.Column]:
    """
    Utility function for creating child primary key in joined table inheritance.
    """
    for base in cls.__bases__:
        # TODO: after dropped Python 3.7
        # if table_name := getattr(base, '__tablename__'):
        if hasattr(base, '__tablename__'):
            table_name = getattr(base, '__tablename__')
            foreign_key = sa.ForeignKey(f'{table_name}.pk')
            return sa.Column(foreign_key, primary_key=True)
    return None



from sqlalchemy_things.declarative.base import (
    DeclarativeMixin,
    InheritedDeclarativeMixin,
    PolymorphicMixin,
    get_inherited_column,
)
from sqlalchemy_things.declarative.primary_keys import (
    BigIntegerPrimaryKeyMixin,
    CascadeBigIntegerPrimaryKeyMixin,
    CascadeDateTimePrimaryKeyMixin,
    CascadeIntegerPrimaryKeyMixin,
    CascadeUUIDPrimaryKeyMixin,
    DateTimePrimaryKeyMixin,
    IntegerPrimaryKeyMixin,
    ParentPrimaryKeyMixin,
    UUIDPrimaryKeyMixin,
    get_inherited_primary_key,
)

__all__ = [
    'BigIntegerPrimaryKeyMixin',
    'DateTimePrimaryKeyMixin',
    'DeclarativeMixin',
    'CascadeBigIntegerPrimaryKeyMixin',
    'CascadeDateTimePrimaryKeyMixin',
    'InheritedDeclarativeMixin',
    'CascadeIntegerPrimaryKeyMixin',
    'CascadeUUIDPrimaryKeyMixin',
    'ParentPrimaryKeyMixin',
    'IntegerPrimaryKeyMixin',
    'PolymorphicMixin',
    'UUIDPrimaryKeyMixin',
    'get_inherited_column',
    'get_inherited_primary_key',
]



from typing import Any, Iterable, Optional

import sqlalchemy as sa
from sqlalchemy import Table, orm


@orm.declarative_mixin
class DeclarativeMixin:
    __table__: 'Table'


@orm.declarative_mixin
class InheritedDeclarativeMixin(DeclarativeMixin):
    __bases__: Iterable[Any]


@orm.declarative_mixin
class PolymorphicMixin(DeclarativeMixin):
    """Polymorphic mixin for single table and joined table inheritance."""
    discriminator = sa.Column(sa.String(40), nullable=False)

    __mapper_args__ = {
        'polymorphic_on': discriminator,
    }


def get_inherited_column(
    cls: Any,
    name: str,
    default: sa.Column,
) -> Optional[sa.Column]:
    """
    Utility function for column inheriting in mixins.

    :param name: column name of parent class;
    :param default: default value if not found in parent class.
    """
    if hasattr(cls, '__table__'):
        return cls.__table__.c.get(name, default)
    return None



from typing import Optional, Union

from sqlalchemy.engine.result import ScalarResult
from sqlalchemy.ext.asyncio import AsyncScalarResult, AsyncSession
from sqlalchemy.orm.session import Session
from sqlalchemy.sql import Select, select
from sqlalchemy.sql.functions import count


class OffsetPage:
    def __init__(
        self,
        items: Union[AsyncScalarResult, ScalarResult],
        next_number: Optional[int],
        previous_number: Optional[int],
        total_items: int,
    ):
        self.items = items
        self.next_number = next_number
        self.previous_number = previous_number
        self.total_items = total_items


class OffsetPaginator:
    def __init__(self, page_size: int = 10, max_page: Optional[int] = None):
        self.page_size = page_size
        self.max_page = max_page

    async def get_page_async(
        self,
        session: AsyncSession,
        stmt: Select,
        page_number: int,
    ) -> OffsetPage:
        stmt = self.handle_max_page_attribute(stmt, page_number)
        total_items = (await session.execute(
            select(count()).select_from(stmt.subquery())
        )).scalar_one()

        stmt = stmt.limit(self.page_size)
        stmt = stmt.offset((page_number - 1) * self.page_size)
        items = (await session.execute(stmt)).scalars()

        return self.prepare_page_instance(page_number, total_items, items)

    def get_page_sync(
        self,
        session: Session,
        stmt: Select,
        page_number: int,
    ) -> OffsetPage:
        stmt = self.handle_max_page_attribute(stmt, page_number)
        total_items = session.execute(
            select(count()).select_from(stmt.subquery())
        ).scalar_one()

        stmt = stmt.limit(self.page_size)
        stmt = stmt.offset((page_number - 1) * self.page_size)
        items = session.execute(stmt).scalars()

        return self.prepare_page_instance(page_number, total_items, items)

    def handle_max_page_attribute(
        self,
        stmt: Select,
        page_number: int,
    ) -> Select:
        if self.max_page:
            if page_number > self.max_page:
                raise ValueError

            if self.max_page:
                stmt = stmt.limit(self.max_page * self.page_size)

        return stmt

    def prepare_page_instance(
        self,
        page_number: int,
        total_items: int,
        items: AsyncScalarResult,
    ) -> OffsetPage:
        previous_number = None
        if page_number > 1:
            previous_number = page_number - 1

        next_number = None
        if page_number * self.page_size < total_items:
            next_number = page_number + 1

        return OffsetPage(
            items=items,
            next_number=next_number,
            previous_number=previous_number,
            total_items=total_items,
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sqlalchemy_things',
 'sqlalchemy_things.column_types',
 'sqlalchemy_things.declarative']

package_data = \
{'': ['*']}

install_requires = \
['sqlalchemy>=1.4.21,<2.0.0']

extras_require = \
{'mysql': ['aiomysql>=0.0.21'],
 'postgresql': ['asyncpg>=0.23.0'],
 'sqlite': ['aiosqlite>=0.17.0']}

setup_kwargs = {
    'name': 'sqlalchemy-things',
    'version': '0.6.1',
    'description': 'Utility collection for development with SQLalchemy ORM.',
    'long_description': "=================\nsqlalchemy-things\n=================\n|ReadTheDocs| |PyPI release| |PyPI downloads| |License| |Python versions| |GitHub CI| |Codecov|\n\n.. |ReadTheDocs| image:: https://readthedocs.org/projects/sqlalchemy-things/badge/?version=latest\n  :target: https://sqlalchemy-things.readthedocs.io/en/latest/?badge=latest\n  :alt: Read The Docs build\n\n.. |PyPI release| image:: https://badge.fury.io/py/sqlalchemy-things.svg\n  :target: https://pypi.org/project/sqlalchemy-things/\n  :alt: Release\n\n.. |PyPI downloads| image:: https://img.shields.io/pypi/dm/sqlalchemy-things\n  :target: https://pypistats.org/packages/sqlalchemy-things\n  :alt: PyPI downloads count\n\n.. |License| image:: https://img.shields.io/badge/License-MIT-green\n  :target: https://github.com/ri-gilfanov/sqlalchemy-things/blob/master/LICENSE\n  :alt: MIT License\n\n.. |Python versions| image:: https://img.shields.io/badge/Python-3.7%20%7C%203.8%20%7C%203.9-blue\n  :target: https://pypi.org/project/sqlalchemy-things/\n  :alt: Python version support\n\n.. |GitHub CI| image:: https://github.com/ri-gilfanov/sqlalchemy-things/actions/workflows/ci.yml/badge.svg?branch=master\n  :target: https://github.com/ri-gilfanov/sqlalchemy-things/actions/workflows/ci.yml\n  :alt: GitHub continuous integration\n\n.. |Codecov| image:: https://codecov.io/gh/ri-gilfanov/sqlalchemy-things/branch/master/graph/badge.svg\n  :target: https://codecov.io/gh/ri-gilfanov/sqlalchemy-things\n  :alt: codecov.io status for master branch\n\nUtility collection for development with `SQLAlchemy 1.4 / 2.0\n<https://www.sqlalchemy.org/>`_ ORM.\n\nDocumentation\n-------------\nhttps://sqlalchemy-things.readthedocs.io\n\nInstallation\n------------\nInstalling ``sqlalchemy-things`` with pip: ::\n\n  pip install sqlalchemy-things\n\nExamples\n--------\nSingle table inheritance\n^^^^^^^^^^^^^^^^^^^^^^^^\n.. code-block:: python\n\n  from sqlalchemy_things.declarative import (\n      IntegerPrimaryKeyMixin,\n      PolymorphicMixin,\n  )\n\n  metadata = sa.MetaData()\n  Base = orm.declarative_base(metadata=metadata)\n\n\n  class Parent(Base, IntegerPrimaryKeyMixin, PolymorphicMixin):\n      __tablename__ = 'single_table'\n\n\n  class ChildA(Parent):\n      __mapper_args__ = {'polymorphic_identity': 'child_a'}\n      some_field = sa.Column(sa.String(255))\n\n\n  class ChildB(Parent):\n      __mapper_args__ = {'polymorphic_identity': 'child_b'}\n      other_filed = sa.Column(sa.String(127))\n\nJoined table inheritance with cascade primary key mixins\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n.. code-block:: python\n\n  from sqlalchemy_things.declarative import (\n      CascadeIntegerPrimaryKeyMixin,\n      PolymorphicMixin,\n  )\n\n  metadata = sa.MetaData()\n  Base = orm.declarative_base(metadata=metadata)\n\n\n  class Parent(Base, CascadeIntegerPrimaryKeyMixin, PolymorphicMixin):\n      __tablename__ = 'cascade_pk_parent_table'\n\n\n  class ChildA(Parent):\n      __tablename__ = 'cascade_pk_child_table_a'\n      __mapper_args__ = {'polymorphic_identity': 'child_a'}\n      some_field = sa.Column(sa.String(255))\n\n\n  class ChildB(Parent):\n      __tablename__ = 'cascade_pk_child_table_b'\n      __mapper_args__ = {'polymorphic_identity': 'child_b'}\n      some_field = sa.Column(sa.String(127))\n\n\nJoined table inheritance with simple primary key mixins\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n.. code-block:: python\n\n  from sqlalchemy_things.declarative import (\n      IntegerPrimaryKeyMixin,\n      ParentPrimaryKeyMixin,\n      PolymorphicMixin,\n  )\n\n  metadata = sa.MetaData()\n  Base = orm.declarative_base(metadata=metadata)\n\n\n  class Parent(Base, IntegerPrimaryKeyMixin, PolymorphicMixin):\n      __tablename__ = 'inherited_pk_parent_table'\n\n\n  class ChildA(ParentPrimaryKeyMixin, Parent):\n      __tablename__ = 'inherited_pk_child_table_a'\n      __mapper_args__ = {'polymorphic_identity': 'child_a'}\n      some_field = sa.Column(sa.String(255))\n\n\n  class ChildB(ParentPrimaryKeyMixin, Parent):\n      __tablename__ = 'inherited_pk_child_table_b'\n      __mapper_args__ = {'polymorphic_identity': 'child_b'}\n      some_field = sa.Column(sa.String(127))\n",
    'author': 'Ruslan Ilyasovich Gilfanov',
    'author_email': 'ri.gilfanov@yandex.ru',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ri-gilfanov/sqlalchemy-things',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
